//! Vertical floating-point math operations.
#![allow(clippy::useless_transmute)]

#[macro_use]
crate mod macros;
crate mod abs;
crate mod cos;
crate mod cos_pi;
crate mod exp;
crate mod ln;
crate mod mul_add;
crate mod mul_adde;
crate mod powf;
crate mod sin;
crate mod sin_cos_pi;
crate mod sin_pi;
crate mod sqrt;
crate mod sqrte;
crate mod tanh;
