//! Implementation of the `ops` traits
#[macro_use]
mod vector_mask_bitwise;
#[macro_use]
mod scalar_mask_bitwise;

#[macro_use]
mod vector_arithmetic;
#[macro_use]
mod scalar_arithmetic;

#[macro_use]
mod vector_bitwise;
#[macro_use]
mod scalar_bitwise;

#[macro_use]
mod vector_shifts;
#[macro_use]
mod scalar_shifts;

#[macro_use]
mod vector_rotates;

#[macro_use]
mod vector_neg;

#[macro_use]
mod vector_int_min_max;

#[macro_use]
mod vector_float_min_max;
